# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Samragni Banerjee <samragnibanerjee4@gmail.com>
#         Alexander Sokolov <alexander.y.sokolov@gmail.com>
#
import time
import numpy as np
import pyscf.ao2mo as ao2mo
import pyscf.adc
import pyscf.adc.radc
from pyscf.adc import radc_ao2mo
import itertools

from itertools import product
from pyscf import lib
from pyscf.pbc import scf
from pyscf.pbc import df
from pyscf.pbc import mp
from pyscf.lib import logger
from pyscf.pbc.adc import kadc_rhf
from pyscf.pbc.adc import kadc_ao2mo
from pyscf.pbc.adc import dfadc
from pyscf import __config__
from pyscf.pbc.mp.kmp2 import (get_nocc, get_nmo, padding_k_idx,_padding_k_idx,
                               padded_mo_coeff, get_frozen_mask, _add_padding)
from pyscf.pbc.cc.kccsd_rhf import _get_epq
from pyscf.pbc.cc.kccsd_t_rhf import _get_epqr
from pyscf.pbc.lib import kpts_helper
from pyscf.lib.parameters import LOOSE_ZERO_TOL, LARGE_DENOM  # noqa

from pyscf.pbc import tools
import h5py
import tempfile


def vector_size(adc):

    nkpts = adc.nkpts
    nocc = adc.nocc
    nvir = adc.nmo - adc.nocc

    n_singles = nvir
    n_doubles = nkpts * nkpts * nocc * nvir * nvir
    size = n_singles + n_doubles

    return size


def get_imds(adc, eris=None):

    cput0 = (time.process_time(), time.time())
    log = logger.Logger(adc.stdout, adc.verbose)

    if adc.method not in ("adc(2)", "adc(2)-x", "adc(3)"):
        raise NotImplementedError(adc.method)

    method = adc.method
    nkpts = adc.nkpts

    mo_energy =  adc.mo_energy
    mo_coeff =  adc.mo_coeff
    nocc = adc.nocc
    nvir = adc.nmo - adc.nocc
    kconserv = adc.khelper.kconserv

    mo_coeff, mo_energy = _add_padding(adc, mo_coeff, mo_energy)

    e_occ = [mo_energy[k][:nocc] for k in range(nkpts)]
    e_vir = [mo_energy[k][nocc:] for k in range(nkpts)]

    e_occ = np.array(e_occ)
    e_vir = np.array(e_vir)

    idn_vir = np.identity(nvir)
    if eris is None:
        eris = adc.transform_integrals()

    eris_ovov = eris.ovov

    # a-b block
    # Zeroth-order terms
    M_ab = np.empty((nkpts,nvir,nvir),dtype=mo_coeff.dtype)

    for ka in range(nkpts):
        kb = ka
        M_ab[ka] = lib.einsum('ab,a->ab', idn_vir, e_vir[ka])
        for kl in range(nkpts):
            for km in range(nkpts):

                kd = kconserv[kl,ka,km]
                # Second-order terms
                t2_1_mla = adc.t2[0][km,kl,ka]
                M_ab[ka] += 0.5 * 0.5 * \
                    lib.einsum('mlad,lbmd->ab',t2_1_mla, eris_ovov[kl,kb,km],optimize=True)
                M_ab[ka] -= 0.5 * 0.5 * \
                    lib.einsum('mlad,ldmb->ab',t2_1_mla, eris_ovov[kl,kd,km],optimize=True)

                t2_1_lma = adc.t2[0][kl,km,ka]
                M_ab[ka] -= 0.5 * 0.5 * \
                    lib.einsum('lmad,lbmd->ab',t2_1_lma, eris_ovov[kl,kb,km],optimize=True)
                M_ab[ka] -= 0.5 * \
                    lib.einsum('lmad,lbmd->ab',t2_1_lma, eris_ovov[kl,kb,km],optimize=True)
                M_ab[ka] += 0.5 * 0.5 * \
                    lib.einsum('lmad,ldmb->ab',t2_1_lma, eris_ovov[kl,kd,km],optimize=True)
                del t2_1_lma

                t2_1_lmb = adc.t2[0][kl,km,kb]
                M_ab[ka] -= 0.5 * 0.5 * \
                    lib.einsum('lmbd,lamd->ab',t2_1_lmb.conj(),
                               eris_ovov[kl,ka,km].conj(),optimize=True)
                M_ab[ka] -= 0.5 * \
                    lib.einsum('lmbd,lamd->ab',t2_1_lmb.conj(),
                               eris_ovov[kl,ka,km].conj(),optimize=True)
                M_ab[ka] += 0.5 * 0.5 * \
                    lib.einsum('lmbd,ldma->ab',t2_1_lmb.conj(),
                               eris_ovov[kl,kd,km].conj(),optimize=True)
                del t2_1_lmb

                t2_1_mlb = adc.t2[0][km,kl,kb]
                M_ab[ka] += 0.5 * 0.5 * \
                    lib.einsum('mlbd,lamd->ab',t2_1_mlb.conj(),
                               eris_ovov[kl,ka,km].conj(),optimize=True)
                M_ab[ka] -= 0.5 * 0.5 * \
                    lib.einsum('mlbd,ldma->ab',t2_1_mlb.conj(),
                               eris_ovov[kl,kd,km].conj(),optimize=True)
                del t2_1_mlb

    cput0 = log.timer_debug1("Completed M_ab second-order terms ADC(2) calculation", *cput0)

    if(method =='adc(3)'):

        eris_oovv = eris.oovv
        eris_ovvo = eris.ovvo

        t1_2 = adc.t1[0]

        for ka in range(nkpts):
            kb = ka
            for kl in range(nkpts):

                if isinstance(eris.ovvv, type(None)):
                    chnk_size = adc.chnk_size
                    if chnk_size > nocc:
                        chnk_size = nocc
                    a = 0
                    for p in range(0,nocc,chnk_size):
                        kd = kconserv[ka,kb,kl]
                        eris_ovvv = dfadc.get_ovvv_df(
                            adc, eris.Lov[kl,kd], eris.Lvv[ka,kb], p, chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                        k = eris_ovvv.shape[0]
                        M_ab[ka] += 2. * lib.einsum('ld,ldab->ab',t1_2[kl]
                                                    [a:a+k], eris_ovvv,optimize=True)
                        del eris_ovvv
                        kd = kconserv[kb,ka,kl]
                        eris_ovvv = dfadc.get_ovvv_df(
                            adc, eris.Lov[kl,kd], eris.Lvv[kb,ka], p, chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                        k = eris_ovvv.shape[0]
                        M_ab[ka] += 2. * lib.einsum('ld,ldba->ab',t1_2[kl]
                                                    [a:a+k].conj(), eris_ovvv.conj(),optimize=True)
                        del eris_ovvv
                        kd = kconserv[ka,kb,kl]
                        eris_ovvv = dfadc.get_ovvv_df(
                            adc, eris.Lov[kl,kb], eris.Lvv[ka,kd], p, chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                        k = eris_ovvv.shape[0]
                        M_ab[ka] -= lib.einsum('ld,lbad->ab',t1_2[kl][a:a+k],
                                               eris_ovvv,optimize=True)
                        del eris_ovvv
                        kd = kconserv[kb,ka,kl]
                        eris_ovvv = dfadc.get_ovvv_df(
                            adc, eris.Lov[kl,ka], eris.Lvv[kb,kd], p, chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                        k = eris_ovvv.shape[0]
                        M_ab[ka] -= lib.einsum('ld,labd->ab',t1_2[kl,a:a+k].conj(),
                                               eris_ovvv.conj(),optimize=True)
                        del eris_ovvv
                        a += k

                else :
                    eris_ovvv = eris.ovvv
                    kd = kconserv[ka,kb,kl]
                    M_ab[ka] += 2. * lib.einsum('ld,ldab->ab',t1_2[kl],
                                                eris_ovvv[kl,kd,ka],optimize=True)
                    M_ab[ka] -=  lib.einsum('ld,lbad->ab',t1_2[kl],
                                            eris_ovvv[kl,kb,ka],optimize=True)
                    kd = kconserv[kb,ka,kl]
                    M_ab[ka] += 2. * lib.einsum('ld,ldba->ab',t1_2[kl].conj(),
                                                eris_ovvv[kl,kd,kb].conj(),optimize=True)
                    M_ab[ka] -= lib.einsum('ld,labd->ab',t1_2[kl].conj(),
                                           eris_ovvv[kl,ka,kb].conj(),optimize=True)

        cput0 = log.timer_debug1("Completed M_ab ovvv ADC(3) calculation", *cput0)

        for ka in range(nkpts):
            kb = ka
            for kl in range(nkpts):
                for km in range(nkpts):
                    kd = kconserv[km,ka,kl]

                    t2_1 = adc.t2[0]
                    t2_2_lma = adc.t2[1][kl,km,ka]
                    M_ab[ka] -= 0.5 * 0.5 * \
                        lib.einsum('lmad,lbmd->ab',t2_2_lma, eris_ovov[kl,kb,km],optimize=True)
                    M_ab[ka] += 0.5 * 0.5 * \
                        lib.einsum('lmad,ldmb->ab',t2_2_lma, eris_ovov[kl,kd,km],optimize=True)
                    M_ab[ka] -= 0.5 *    lib.einsum('lmad,lbmd->ab',
                                                    t2_2_lma, eris_ovov[kl,kb,km],optimize=True)

                    t2_2_mla = adc.t2[1][km,kl,ka]
                    M_ab[ka] += 0.5 * 0.5 * \
                        lib.einsum('mlad,lbmd->ab',t2_2_mla, eris_ovov[kl,kb,km],optimize=True)
                    M_ab[ka] -= 0.5 * 0.5 * \
                        lib.einsum('mlad,ldmb->ab',t2_2_mla, eris_ovov[kl,kd,km],optimize=True)

                    t2_2_lmb = adc.t2[1][kl,km,kb]
                    M_ab[ka] -= 0.5 * 0.5 * \
                        lib.einsum('lmbd,lamd->ab',t2_2_lmb.conj(),
                                   eris_ovov[kl,ka,km].conj(),optimize=True)
                    M_ab[ka] += 0.5 * 0.5 * \
                        lib.einsum('lmbd,ldma->ab',t2_2_lmb.conj(),
                                   eris_ovov[kl,kd,km].conj(),optimize=True)
                    M_ab[ka] -= 0.5 * lib.einsum('lmbd,lamd->ab',t2_2_lmb.conj(),
                                                 eris_ovov[kl,ka,km].conj(),optimize=True)

                    t2_2_mlb = adc.t2[1][km,kl,kb]
                    M_ab[ka] += 0.5 * 0.5 * \
                        lib.einsum('mlbd,mdla->ab',t2_2_mlb.conj(),
                                   eris_ovov[km,kd,kl].conj(),optimize=True)
                    M_ab[ka] -= 0.5 * 0.5 * \
                        lib.einsum('mlbd,ldma->ab',t2_2_mlb.conj(),
                                   eris_ovov[kl,kd,km].conj(),optimize=True)
                    del t2_2_mlb

            log.timer_debug1("Starting the small integrals  calculation")
            for kn, ke, kd in kpts_helper.loop_kkk(nkpts):

                kl = kconserv[ke,kn,kd]
                km = kconserv[kb,kl,kd]
                temp_t2_v_1 = lib.einsum(
                    'lned,mlbd->nemb',t2_1[kl,kn,ke], t2_1[km,kl,kb].conj(),optimize=True)
                M_ab[ka] -= 0.5 *     lib.einsum('nemb,nmae->ab',
                                                 temp_t2_v_1, eris_oovv[kn,km,ka], optimize=True)
                M_ab[ka] += 0.5 *2. * lib.einsum('nemb,neam->ab',
                                                 temp_t2_v_1, eris_ovvo[kn,ke,ka], optimize=True)
                M_ab[ka] += 0.5 *2. * lib.einsum('nema,nebm->ab',
                                                 temp_t2_v_1, eris_ovvo[kn,ke,kb], optimize=True)
                M_ab[ka] -= 0.5 *     lib.einsum('nema,nmbe->ab',
                                                 temp_t2_v_1, eris_oovv[kn,km,kb], optimize=True)
                del temp_t2_v_1

                temp_t2_v_1 = lib.einsum(
                    'nled,lmbd->mbne',t2_1[kn,kl,ke], t2_1[kl,km,kb].conj(),optimize=True)
                M_ab[ka] -= 0.5 *     lib.einsum('mbne,nmae->ab',
                                                 temp_t2_v_1, eris_oovv[kn,km,ka], optimize=True)
                M_ab[ka] += 0.5 *2. * lib.einsum('mbne,neam->ab',
                                                 temp_t2_v_1, eris_ovvo[kn,ke,ka], optimize=True)
                M_ab[ka] -= 0.5 *     lib.einsum('mane,nmbe->ab',
                                                 temp_t2_v_1, eris_oovv[kn,km,kb], optimize=True)
                M_ab[ka] += 0.5 *2. * lib.einsum('mane,nebm->ab',
                                                 temp_t2_v_1, eris_ovvo[kn,ke,kb], optimize=True)
                del temp_t2_v_1

                temp_t2_v_2 = lib.einsum(
                    'nled,mlbd->nemb',t2_1[kn,kl,ke], t2_1[km,kl,kb].conj(),optimize=True)
                M_ab[ka] += 0.5 * 2. * lib.einsum('nemb,nmae->ab',
                                                  temp_t2_v_2, eris_oovv[kn,km,ka], optimize=True)
                M_ab[ka] -= 0.5 * 4. * lib.einsum('nemb,neam->ab',
                                                  temp_t2_v_2, eris_ovvo[kn,ke,ka], optimize=True)
                M_ab[ka] += 0.5 * 2. * lib.einsum('nema,nmbe->ab',
                                                  temp_t2_v_2, eris_oovv[kn,km,kb], optimize=True)
                M_ab[ka] -= 0.5 * 4. * lib.einsum('nema,nebm->ab',
                                                  temp_t2_v_2, eris_ovvo[kn,ke,kb], optimize=True)
                del temp_t2_v_2

                temp_t2_v_3 = lib.einsum(
                    'lned,lmbd->nemb',t2_1[kl,kn,ke], t2_1[kl,km,kb].conj(),optimize=True)
                M_ab[ka] -= 0.5 *      lib.einsum('nemb,neam->ab',
                                                  temp_t2_v_3, eris_ovvo[kn,ke,ka], optimize=True)
                M_ab[ka] += 0.5 * 2. * lib.einsum('nemb,nmae->ab',
                                                  temp_t2_v_3, eris_oovv[kn,km,ka], optimize=True)
                M_ab[ka] += 0.5 * 2. * lib.einsum('nema,nmbe->ab',
                                                  temp_t2_v_3, eris_oovv[kn,km,kb], optimize=True)
                M_ab[ka] -= 0.5 *      lib.einsum('nema,nebm->ab',
                                                  temp_t2_v_3, eris_ovvo[kn,ke,kb], optimize=True)
                del temp_t2_v_3

                kl = kconserv[ke,kn,kd]
                km = kconserv[ke,kl,kd]
                temp_t2_v_8 = lib.einsum(
                    'lned,mled->mn',t2_1[kl,kn,ke], t2_1[km,kl,ke].conj(),optimize=True)
                M_ab[ka] += 2.* lib.einsum('mn,nmab->ab',temp_t2_v_8,
                                           eris_oovv[kn,km,ka], optimize=True)
                M_ab[ka] -= lib.einsum('mn,nbam->ab', temp_t2_v_8,
                                       eris_ovvo[kn,kb,ka], optimize=True)
                del temp_t2_v_8

                temp_t2_v_9 = lib.einsum(
                    'nled,mled->mn',t2_1[kn,kl,ke], t2_1[km,kl,ke].conj(),optimize=True)
                M_ab[ka] -= 4. * lib.einsum('mn,nmab->ab',temp_t2_v_9,
                                            eris_oovv[kn,km,ka], optimize=True)
                M_ab[ka] += 2. * lib.einsum('mn,nbam->ab',temp_t2_v_9,
                                            eris_ovvo[kn,kb,ka], optimize=True)
                del temp_t2_v_9

            for km in range(nkpts):
                for kl in range(nkpts):
                    kf = kconserv[km,kl,ka]
                    temp_t2 = adc.imds.t2_1_vvvv[:]
                    t2_1_mla = adc.t2[0][km,kl,ka]
                    M_ab[ka] -= 0.5 * 0.25* \
                        lib.einsum('mlaf,mlbf->ab',t2_1_mla,
                                   temp_t2[km,kl,kb].conj(), optimize=True)
                    M_ab[ka] += 0.5 * 0.25* \
                        lib.einsum('mlaf,lmbf->ab',t2_1_mla,
                                   temp_t2[kl,km,kb].conj(), optimize=True)
                    M_ab[ka] += 0.5 * 0.25* \
                        lib.einsum('mlaf,mlfb->ab',t2_1_mla,
                                   temp_t2[km,kl,kf].conj(), optimize=True)
                    M_ab[ka] -= 0.5 * 0.25* \
                        lib.einsum('mlaf,lmfb->ab',t2_1_mla,
                                   temp_t2[kl,km,kf].conj(), optimize=True)
                    M_ab[ka] -= 0.5 *      lib.einsum('mlaf,mlbf->ab',
                                                      t2_1_mla, temp_t2[km,kl,kb].conj(), optimize=True)
                    del t2_1_mla

                    t2_1_lma = adc.t2[0][kl,km,ka]
                    M_ab[ka] += 0.5 * 0.25* \
                        lib.einsum('lmaf,mlbf->ab',t2_1_lma,
                                   temp_t2[km,kl,kb].conj(), optimize=True)
                    M_ab[ka] -= 0.5 * 0.25* \
                        lib.einsum('lmaf,lmbf->ab',t2_1_lma,
                                   temp_t2[kl,km,kb].conj(), optimize=True)
                    M_ab[ka] -= 0.5 * 0.25* \
                        lib.einsum('lmaf,mlfb->ab',t2_1_lma,
                                   temp_t2[km,kl,kf].conj(), optimize=True)
                    M_ab[ka] += 0.5 * 0.25* \
                        lib.einsum('lmaf,lmfb->ab',t2_1_lma,
                                   temp_t2[kl,km,kf].conj(), optimize=True)
                    del t2_1_lma

                    kd = kconserv[km,ka,kl]
                    t2_1_mlb = adc.t2[0][km,kl,kb]
                    M_ab[ka] -= 0.5 * 0.25* \
                        lib.einsum('mlad,mlbd->ab', temp_t2[km,
                                   kl,ka].conj(), t2_1_mlb, optimize=True)
                    M_ab[ka] += 0.5 * 0.25* \
                        lib.einsum('lmad,mlbd->ab', temp_t2[kl,
                                   km,ka].conj(), t2_1_mlb, optimize=True)
                    M_ab[ka] -= 0.5 *      lib.einsum('mlad,mlbd->ab',
                                                      temp_t2[km,kl,ka].conj(), t2_1_mlb, optimize=True)

                    M_ab[ka] += 0.5 * 0.25* \
                        lib.einsum('lmad,mlbd->ab', temp_t2[kl,
                                   km,ka].conj(), t2_1_mlb, optimize=True)
                    M_ab[ka] -= 0.5 * 0.25* \
                        lib.einsum('mlad,mlbd->ab', temp_t2[km,
                                   kl,ka].conj(), t2_1_mlb, optimize=True)
                    del t2_1_mlb

                    t2_1_lmb = adc.t2[0][kl,km,kb]
                    M_ab[ka] += 0.5 * 0.25* \
                        lib.einsum('mlad,lmbd->ab', temp_t2[km,
                                   kl,ka].conj(), t2_1_lmb, optimize=True)
                    M_ab[ka] -= 0.5 * 0.25* \
                        lib.einsum('lmad,lmbd->ab', temp_t2[kl,
                                   km,ka].conj(), t2_1_lmb, optimize=True)

                    M_ab[ka] -= 0.5 * 0.25* \
                        lib.einsum('lmad,lmbd->ab', temp_t2[kl,
                                   km,ka].conj(), t2_1_lmb, optimize=True)
                    M_ab[ka] += 0.5 * 0.25* \
                        lib.einsum('mlad,lmbd->ab', temp_t2[km,
                                   kl,ka].conj(), t2_1_lmb, optimize=True)
                    del temp_t2
                    del t2_1_lmb

                    for kd in range(nkpts):
                        kf = kconserv[km,kd,kl]
                        ke = kconserv[kb,ka,kf]
                        if isinstance(eris.vvvv, type(None)):
                            chnk_size = adc.chnk_size
                            if chnk_size > nvir:
                                chnk_size = nvir
                            a = 0
                            for p in range(0,nvir,chnk_size):
                                eris_vvvv = dfadc.get_vvvv_df(
                                    adc, eris.Lvv[kb,ka], eris.Lvv[ke,kf], p, chnk_size)/nkpts
                                k = eris_vvvv.shape[0]
                                M_ab[ka,a:a+k] -= lib.einsum('mldf,mled,aebf->ab',t2_1[km,kl,kd],
                                                             t2_1[km,kl,ke].conj(), eris_vvvv, optimize=True)
                                M_ab[ka,a:a+k] += lib.einsum('mldf,lmed,aebf->ab',t2_1[km,kl,kd],
                                                             t2_1[kl,km,ke].conj(), eris_vvvv, optimize=True)
                                M_ab[ka,a:a+k] += lib.einsum('lmdf,mled,aebf->ab',t2_1[kl,km,kd],
                                                             t2_1[km,kl,ke].conj(), eris_vvvv, optimize=True)
                                M_ab[ka,a:a+k] -= lib.einsum('lmdf,lmed,aebf->ab',t2_1[kl,km,kd],
                                                             t2_1[kl,km,ke].conj(), eris_vvvv, optimize=True)
                                M_ab[ka,a:a+k] += 2.*lib.einsum(
                                    'mlfd,mled,aebf->ab',t2_1[km,kl,kf], t2_1[km,kl,ke].conj(), eris_vvvv,
                                    optimize=True)
                                del eris_vvvv

                                eris_vvvv = dfadc.get_vvvv_df(
                                    adc, eris.Lvv[kb,kf], eris.Lvv[ke,ka], p, chnk_size)/nkpts
                                M_ab[ka,a:a+k] += 0.5*lib.einsum(
                                    'mldf,mled,aefb->ab',t2_1[km,kl,kd], t2_1[km,kl,ke].conj(), eris_vvvv,
                                    optimize=True)
                                M_ab[ka,a:a+k] -= 0.5*lib.einsum(
                                    'mldf,lmed,aefb->ab',t2_1[km,kl,kd], t2_1[kl,km,ke].conj(), eris_vvvv,
                                    optimize=True)
                                M_ab[ka,a:a+k] -= 0.5*lib.einsum(
                                    'lmdf,mled,aefb->ab',t2_1[kl,km,kd], t2_1[km,kl,ke].conj(), eris_vvvv,
                                    optimize=True)
                                M_ab[ka,a:a+k] += 0.5*lib.einsum(
                                    'lmdf,lmed,aefb->ab',t2_1[kl,km,kd], t2_1[kl,km,ke].conj(), eris_vvvv,
                                    optimize=True)
                                M_ab[ka,a:a+k] -= lib.einsum('mlfd,mled,aefb->ab',t2_1[km,kl,kf],
                                                             t2_1[km,kl,ke].conj(), eris_vvvv, optimize=True)
                                del eris_vvvv
                                a += k

                        elif isinstance(eris.vvvv, np.ndarray):
                            eris_vvvv =  eris.vvvv
                            M_ab[ka] -= lib.einsum('mldf,mled,aebf->ab',t2_1[km,kl,kd],
                                                   t2_1[km,kl,ke].conj(), eris_vvvv[ka,ke,kb], optimize=True)
                            M_ab[ka] += lib.einsum('mldf,lmed,aebf->ab',t2_1[km,kl,kd],
                                                   t2_1[kl,km,ke].conj(), eris_vvvv[ka,ke,kb], optimize=True)
                            M_ab[ka] += lib.einsum('lmdf,mled,aebf->ab',t2_1[kl,km,kd],
                                                   t2_1[km,kl,ke].conj(), eris_vvvv[ka,ke,kb], optimize=True)
                            M_ab[ka] -= lib.einsum('lmdf,lmed,aebf->ab',t2_1[kl,km,kd],
                                                   t2_1[kl,km,ke].conj(), eris_vvvv[ka,ke,kb], optimize=True)
                            M_ab[ka] += 2.*lib.einsum('mlfd,mled,aebf->ab',t2_1[km,kl,kf],
                                                      t2_1[km,kl,ke].conj(), eris_vvvv[ka,ke,kb], optimize=True)
                            M_ab[ka] += 0.5*lib.einsum('mldf,mled,aefb->ab',t2_1[km,kl,kd],
                                                       t2_1[km,kl,ke].conj(), eris_vvvv[ka,ke,kf], optimize=True)
                            M_ab[ka] -= 0.5*lib.einsum('mldf,lmed,aefb->ab',t2_1[km,kl,kd],
                                                       t2_1[kl,km,ke].conj(), eris_vvvv[ka,ke,kf], optimize=True)
                            M_ab[ka] -= 0.5*lib.einsum('lmdf,mled,aefb->ab',t2_1[kl,km,kd],
                                                       t2_1[km,kl,ke].conj(), eris_vvvv[ka,ke,kf], optimize=True)
                            M_ab[ka] += 0.5*lib.einsum('lmdf,lmed,aefb->ab',t2_1[kl,km,kd],
                                                       t2_1[kl,km,ke].conj(), eris_vvvv[ka,ke,kf], optimize=True)
                            M_ab[ka] -= lib.einsum('mlfd,mled,aefb->ab',t2_1[km,kl,kf],
                                                   t2_1[km,kl,ke].conj(), eris_vvvv[ka,ke,kf], optimize=True)
                        else :
                            chnk_size = adc.chnk_size
                            if chnk_size > nvir:
                                chnk_size = nvir
                            a = 0
                            for p in range(0,nvir,chnk_size):
                                eris_vvvv = eris.vvvv[kb,ke,ka,p:p+chnk_size]
                                k = eris_vvvv.shape[0]
                                M_ab[ka,a:a+k] -= lib.einsum('mldf,mled,aebf->ab',t2_1[km,kl,kd],
                                                             t2_1[km,kl,ke].conj(), eris_vvvv, optimize=True)
                                M_ab[ka,a:a+k] += lib.einsum('mldf,lmed,aebf->ab',t2_1[km,kl,kd],
                                                             t2_1[kl,km,ke].conj(), eris_vvvv, optimize=True)
                                M_ab[ka,a:a+k] += lib.einsum('lmdf,mled,aebf->ab',t2_1[kl,km,kd],
                                                             t2_1[km,kl,ke].conj(), eris_vvvv, optimize=True)
                                M_ab[ka,a:a+k] -= lib.einsum('lmdf,lmed,aebf->ab',t2_1[kl,km,kd],
                                                             t2_1[kl,km,ke].conj(), eris_vvvv, optimize=True)
                                M_ab[ka,a:a+k] += 2.*lib.einsum(
                                    'mlfd,mled,aebf->ab',t2_1[km,kl,kf], t2_1[km,kl,ke].conj(), eris_vvvv,
                                    optimize=True)
                                del eris_vvvv
                                eris_vvvv = eris.vvvv[ka,ke,kf,p:p+chnk_size]
                                M_ab[ka,a:a+k] += 0.5*lib.einsum(
                                    'mldf,mled,aefb->ab',t2_1[km,kl,kd], t2_1[km,kl,ke].conj(), eris_vvvv,
                                    optimize=True)
                                M_ab[ka,a:a+k] -= 0.5*lib.einsum(
                                    'mldf,lmed,aefb->ab',t2_1[km,kl,kd], t2_1[kl,km,ke].conj(), eris_vvvv,
                                    optimize=True)
                                M_ab[ka,a:a+k] -= 0.5*lib.einsum(
                                    'lmdf,mled,aefb->ab',t2_1[kl,km,kd], t2_1[km,kl,ke].conj(), eris_vvvv,
                                    optimize=True)
                                M_ab[ka,a:a+k] += 0.5*lib.einsum(
                                    'lmdf,lmed,aefb->ab',t2_1[kl,km,kd], t2_1[kl,km,ke].conj(), eris_vvvv,
                                    optimize=True)
                                M_ab[ka,a:a+k] -= lib.einsum('mlfd,mled,aefb->ab',t2_1[km,kl,kf],
                                                             t2_1[km,kl,ke].conj(), eris_vvvv, optimize=True)
                                del eris_vvvv
                                a += k

    cput0 = log.timer_debug1("Completed M_ab ADC(3) calculation", *cput0)

    return M_ab


def get_diag(adc,kshift,M_ab=None,eris=None):

    log = logger.Logger(adc.stdout, adc.verbose)

    if adc.method not in ("adc(2)", "adc(2)-x", "adc(3)"):
        raise NotImplementedError(adc.method)

    if M_ab is None:
        M_ab = adc.get_imds()

    nkpts = adc.nkpts
    kconserv = adc.khelper.kconserv

    nocc = adc.nocc
    nvir = adc.nmo - adc.nocc
    n_singles = nvir
    n_doubles = nkpts * nkpts * nocc * nvir * nvir

    dim = n_singles + n_doubles

    s1 = 0
    f1 = n_singles
    s2 = f1
    f2 = s2 + n_doubles

    mo_energy =  adc.mo_energy
    mo_coeff =  adc.mo_coeff
    nocc = adc.nocc
    nmo = adc.nmo
    nvir = nmo - nocc
    mo_coeff, mo_energy = _add_padding(adc, mo_coeff, mo_energy)

    e_occ = [mo_energy[k][:nocc] for k in range(nkpts)]
    e_vir = [mo_energy[k][nocc:] for k in range(nkpts)]

    e_vir = np.array(e_vir)
    e_occ = np.array(e_occ)

    diag = np.zeros((dim), dtype=np.complex128)
    doubles = np.zeros((nkpts,nkpts,nocc*nvir*nvir),dtype=np.complex128)

    # Compute precond in h1-h1 block
    M_ab_diag = np.diagonal(M_ab[kshift])
    diag[s1:f1] = M_ab_diag.copy()

    # Compute precond in 2p1h-2p1h block
    for kj in range(nkpts):
        for ka in range(nkpts):
            kb = kconserv[kshift,ka,kj]
            d_ab = e_vir[ka][:,None] + e_vir[kb]
            d_i = e_occ[kj][:,None]
            D_n = -d_i + d_ab.reshape(-1)
            doubles[kj,ka] += D_n.reshape(-1)

    diag[s2:f2] = doubles.reshape(-1)
    log.timer_debug1("Completed ea_diag calculation")

    return diag


def matvec(adc, kshift, M_ab=None, eris=None):

    if adc.method not in ("adc(2)", "adc(2)-x", "adc(3)"):
        raise NotImplementedError(adc.method)

    method = adc.method

    nkpts = adc.nkpts
    nocc = adc.nocc
    kconserv = adc.khelper.kconserv
    nvir = adc.nmo - adc.nocc
    n_singles = nvir
    n_doubles = nkpts * nkpts * nocc * nvir * nvir

    s_singles = 0
    f_singles = n_singles
    s_doubles = f_singles
    f_doubles = s_doubles + n_doubles

    mo_energy =  adc.mo_energy
    mo_coeff =  adc.mo_coeff
    mo_coeff, mo_energy = _add_padding(adc, mo_coeff, mo_energy)

    e_occ = [mo_energy[k][:nocc] for k in range(nkpts)]
    e_vir = [mo_energy[k][nocc:] for k in range(nkpts)]

    e_vir = np.array(e_vir)
    e_occ = np.array(e_occ)

    if M_ab is None:
        M_ab = adc.get_imds()

    #Calculate sigma vector
    def sigma_(r):
        cput0 = (time.process_time(), time.time())
        log = logger.Logger(adc.stdout, adc.verbose)

        r1 = r[s_singles:f_singles]
        r2 = r[s_doubles:f_doubles]

        r2 = r2.reshape(nkpts,nkpts,nocc,nvir,nvir)
        s2 = np.zeros((nkpts,nkpts,nocc,nvir,nvir), dtype=np.complex128)
        cell = adc.cell
        kpts = adc.kpts
        madelung = tools.madelung(cell, kpts)

############ ADC(2) ij block ############################

        s1 = lib.einsum('ab,b->a',M_ab[kshift],r1)

########### ADC(2) coupling blocks #########################

        for kb in range(nkpts):
            for kc in range(nkpts):
                ki = kconserv[kb,kshift, kc]
                if isinstance(eris.ovvv, type(None)):
                    chnk_size = adc.chnk_size
                    if chnk_size > nocc:
                        chnk_size = nocc
                    a = 0
                    for p in range(0,nocc,chnk_size):
                        eris_ovvv = dfadc.get_ovvv_df(
                            adc, eris.Lov[ki,kc], eris.Lvv[kshift,kb], p, chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                        k = eris_ovvv.shape[0]
                        s1 +=  2. * lib.einsum('icab,ibc->a', eris_ovvv.conj(),
                                               r2[ki,kb,a:a+k], optimize=True)
                        s2[ki,kb,a:a+k] += lib.einsum('icab,a->ibc', eris_ovvv, r1, optimize=True)
                        del eris_ovvv

                        eris_ovvv = dfadc.get_ovvv_df(
                            adc, eris.Lov[ki,kb], eris.Lvv[kshift,kc], p, chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                        s1 -=  lib.einsum('ibac,ibc->a',   eris_ovvv.conj(),
                                          r2[ki,kb,a:a+k], optimize=True)
                        del eris_ovvv
                        a += k
                else :
                    eris_ovvv = eris.ovvv[:]
                    s1 +=  2. * lib.einsum('icab,ibc->a',
                                           eris_ovvv[ki,kc,kshift].conj(), r2[ki,kb], optimize=True)
                    s2[ki,kb] += lib.einsum('icab,a->ibc', eris_ovvv[ki,
                                            kc,kshift], r1, optimize=True)
                    del eris_ovvv


#########    ########## ADC(2) ajk - bil block ############################

                s2[ki, kb] -= lib.einsum('i,ibc->ibc', e_occ[ki], r2[ki, kb])
                s2[ki, kb] += lib.einsum('b,ibc->ibc', e_vir[kb], r2[ki, kb])
                s2[ki, kb] += lib.einsum('c,ibc->ibc', e_vir[kc], r2[ki, kb])

################ ADC(3) ajk - bil block ############################

        if (method == "adc(2)-x" or method == "adc(3)"):

            eris_oovv = eris.oovv
            eris_ovvo = eris.ovvo

            for kx in range(nkpts):
                for ky in range(nkpts):
                    ki = kconserv[ky,kshift, kx]
                    for kz in range(nkpts):
                        kw = kconserv[kx, kz, ky]

                        if isinstance(eris.vvvv, np.ndarray):
                            eris_vvvv = eris.vvvv.reshape(nkpts,nkpts,nkpts,nvir*nvir,nvir*nvir)
                            r2_1 = r2.reshape(nkpts,nkpts,nocc,nvir*nvir)
                            s2[ki, kx] += np.dot(r2_1[ki,kw],eris_vvvv[kx,ky,
                                                 kw].T.conj()).reshape(nocc,nvir,nvir)
                        elif isinstance(eris.vvvv, type(None)):
                            s2[ki,kx] += ea_contract_r_vvvv(adc,r2[ki,kw],eris.Lvv,kx,ky,kw)
                        else :
                            s2[ki,kx] += ea_contract_r_vvvv(adc,r2[ki,kw],eris.vvvv,kx,ky,kw)
                    for kj in range(nkpts):
                        kz = kconserv[ky,ki,kj]
                        s2[ki,kx] -= 0.5*lib.einsum('iyzj,jzx->ixy',
                                                    eris_ovvo[ki,ky,kz],r2[kj,kz],optimize=True)
                        s2[ki,kx] += lib.einsum('iyzj,jxz->ixy',eris_ovvo[ki,
                                                ky,kz],r2[kj,kx],optimize=True)
                        s2[ki,kx] -= 0.5*lib.einsum('ijzy,jxz->ixy',
                                                    eris_oovv[ki,kj,kz],r2[kj,kx],optimize=True)

                        kz = kconserv[kj,ki,kx]
                        s2[ki,kx] -=  0.5*lib.einsum('ijzx,jzy->ixy',
                                                     eris_oovv[ki,kj,kz],r2[kj,kz],optimize=True)

                        kw = kconserv[kj,ki,kx]
                        s2[ki,kx] -=  0.5*lib.einsum('ijwx,jwy->ixy',
                                                     eris_oovv[ki,kj,kw],r2[kj,kw],optimize=True)

                        kw = kconserv[kj,ki,ky]
                        s2[ki,kx] -= 0.5*lib.einsum('ijwy,jxw->ixy',
                                                    eris_oovv[ki,kj,kw],r2[kj,kx],optimize=True)
                        s2[ki,kx] += lib.einsum('iywj,jxw->ixy',eris_ovvo[ki,
                                                ky,kw],r2[kj,kx],optimize=True)
                        s2[ki,kx] -= 0.5*lib.einsum('iywj,jwx->ixy',
                                                    eris_ovvo[ki,ky,kw],r2[kj,kw],optimize=True)

            if adc.exxdiv is not None:
                s2 += -madelung * r2
        if (method == "adc(3)"):

            eris_ovoo = eris.ovoo

################ ADC(3) a - ibc block and ibc-a coupling blocks ########################

            t2_1 = adc.t2[0]

            for kz in range(nkpts):
                for kw in range(nkpts):
                    kj = kconserv[kw, kshift, kz]

                    for kl in range(nkpts):
                        km = kconserv[kw, kl, kz]
                        t2_1_lmz = adc.t2[0][kl,km,kz]
                        ka = kconserv[km, kj, kl]
                        temp_1 =       lib.einsum('lmzw,jzw->jlm',t2_1_lmz,r2[kj,kz])
                        temp = 0.25 * lib.einsum('lmzw,jzw->jlm',t2_1_lmz,r2[kj,kz])
                        temp -= 0.25 * lib.einsum('lmzw,jwz->jlm',t2_1_lmz,r2[kj,kw])
                        del t2_1_lmz

                        t2_1_mlz = adc.t2[0][km,kl,kz]
                        temp -= 0.25 * lib.einsum('mlzw,jzw->jlm',t2_1_mlz,r2[kj,kz])
                        temp += 0.25 * lib.einsum('mlzw,jwz->jlm',t2_1_mlz,r2[kj,kw])
                        del t2_1_mlz

                        s1 += lib.einsum('jlm,lamj->a',temp,   eris_ovoo[kl,ka,km], optimize=True)
                        s1 -= lib.einsum('jlm,malj->a',temp,   eris_ovoo[km,ka,kl], optimize=True)
                        s1 += lib.einsum('jlm,lamj->a',temp_1, eris_ovoo[kl,ka,km], optimize=True)

                        del temp
                        del temp_1

            for kx in range(nkpts):
                for ky in range(nkpts):
                    ki = kconserv[ky, kshift, kx]
                    for kl in range(nkpts):
                        km = kconserv[kx, kl, ky]
                        kb = kconserv[km,ki,kl]
                        temp = lib.einsum(
                            'b,lbmi->lmi',r1,eris_ovoo[kl,kb,km].conj(), optimize=True)
                        s2[ki,kx] += lib.einsum('lmi,lmxy->ixy',temp,
                                                t2_1[kl,km,kx].conj(), optimize=True)
                        del temp

            for kz in range(nkpts):
                for kw in range(nkpts):
                    kj = kconserv[kz, kshift, kw]
                    for kl in range(nkpts):
                        kd = kconserv[kj, kw, kl]
                        t2_1_jlw = adc.t2[0][kj,kl,kw]

                        temp_s_a = lib.einsum('jlwd,jzw->lzd',t2_1_jlw,r2[kj,kz],optimize=True)
                        temp_s_a -= lib.einsum('jlwd,jwz->lzd',t2_1_jlw,r2[kj,kw],optimize=True)

                        temp_t2_r2_1 = lib.einsum('jlwd,jzw->lzd',t2_1_jlw,r2[kj,kz],optimize=True)
                        temp_t2_r2_1 -= lib.einsum('jlwd,jwz->lzd',t2_1_jlw,r2[kj,kw],optimize=True)
                        temp_t2_r2_1 += lib.einsum('jlwd,jzw->lzd',t2_1_jlw,r2[kj,kz],optimize=True)
                        del t2_1_jlw

                        t2_1_ljw = adc.t2[0][kl,kj,kw]
                        temp_s_a -= lib.einsum('ljwd,jzw->lzd',t2_1_ljw,r2[kj,kz],optimize=True)
                        temp_s_a += lib.einsum('ljwd,jwz->lzd',t2_1_ljw,r2[kj,kw],optimize=True)
                        temp_s_a += lib.einsum('ljdw,jzw->lzd',
                                               t2_1[kl,kj,kd],r2[kj,kz],optimize=True)

                        temp_t2_r2_1 -= lib.einsum('ljwd,jzw->lzd',t2_1_ljw,r2[kj,kz],optimize=True)
                        del t2_1_ljw
                        temp_t2_r2_4 = lib.einsum(
                            'jldw,jwz->lzd',t2_1[kj,kl,kd],r2[kj,kw], optimize=True)

                        if isinstance(eris.ovvv, type(None)):
                            chnk_size = adc.chnk_size
                            if chnk_size > nocc:
                                chnk_size = nocc
                            a = 0
                            for p in range(0,nocc,chnk_size):
                                ka = kconserv[kz, kd, kl]
                                eris_ovvv = dfadc.get_ovvv_df(
                                    adc, eris.Lov[kl,kd], eris.Lvv[kz,ka], p,
                                    chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                                k = eris_ovvv.shape[0]

                                s1 += 0.5*lib.einsum('lzd,ldza->a',
                                                     temp_s_a[a:a+k],eris_ovvv,optimize=True)
                                s1 += 0.5*lib.einsum('lzd,ldza->a',
                                                     temp_t2_r2_1[a:a+k],eris_ovvv,optimize=True)
                                del eris_ovvv

                                eris_ovvv = dfadc.get_ovvv_df(
                                    adc, eris.Lov[kl,ka], eris.Lvv[kz,kd], p,
                                    chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                                s1 -= 0.5*lib.einsum('lzd,lazd->a',
                                                     temp_s_a[a:a+k],eris_ovvv,optimize=True)
                                s1 -= 0.5*lib.einsum('lzd,lazd->a',
                                                     temp_t2_r2_4[a:a+k],eris_ovvv,optimize=True)
                                del eris_ovvv
                                a += k
                        else :
                            ka = kconserv[kz, kd, kl]
                            eris_ovvv = eris.ovvv[:]
                            s1 += 0.5*lib.einsum('lzd,ldza->a',temp_s_a,
                                                 eris_ovvv[kl,kd,kz],optimize=True)
                            s1 += 0.5*lib.einsum('lzd,ldza->a',temp_t2_r2_1,
                                                 eris_ovvv[kl,kd,kz],optimize=True)

                            s1 -= 0.5*lib.einsum('lzd,lazd->a',temp_s_a,
                                                 eris_ovvv[kl,ka,kz],optimize=True)
                            s1 -= 0.5*lib.einsum('lzd,lazd->a',temp_t2_r2_4,
                                                 eris_ovvv[kl,ka,kz],optimize=True)
                            del eris_ovvv
                        del temp_s_a
                        del temp_t2_r2_1
                        del temp_t2_r2_4

            for kz in range(nkpts):
                for kw in range(nkpts):
                    kj = kconserv[kz, kshift, kw]
                    for kl in range(nkpts):
                        kd = kconserv[kj, kz, kl]
                        t2_1_jlz = adc.t2[0][kj,kl,kz]

                        temp_s_a_1  =   -lib.einsum('jlzd,jwz->lwd',
                                                    t2_1_jlz,r2[kj,kw],optimize=True)
                        temp_s_a_1 +=    lib.einsum('jlzd,jzw->lwd',
                                                    t2_1_jlz,r2[kj,kz],optimize=True)
                        temp_t2_r2_2  = -lib.einsum('jlzd,jwz->lwd',
                                                    t2_1_jlz,r2[kj,kw],optimize=True)
                        temp_t2_r2_2 +=  lib.einsum('jlzd,jzw->lwd',
                                                    t2_1_jlz,r2[kj,kz],optimize=True)
                        temp_t2_r2_2 -=  lib.einsum('jlzd,jwz->lwd',
                                                    t2_1_jlz,r2[kj,kw],optimize=True)
                        del t2_1_jlz

                        t2_1_ljz = adc.t2[0][kl,kj,kz]
                        temp_s_a_1 += lib.einsum('ljzd,jwz->lwd',t2_1_ljz,r2[kj,kw],optimize=True)
                        temp_s_a_1 -= lib.einsum('ljzd,jzw->lwd',t2_1_ljz,r2[kj,kz],optimize=True)
                        temp_s_a_1 -= lib.einsum('ljdz,jwz->lwd',
                                                 t2_1[kl,kj,kd],r2[kj,kw],optimize=True)

                        temp_t2_r2_2 += lib.einsum('ljzd,jwz->lwd',t2_1_ljz,r2[kj,kw],optimize=True)

                        temp_t2_r2_3 = -lib.einsum('ljzd,jzw->lwd',t2_1_ljz,r2[kj,kz],optimize=True)
                        del t2_1_ljz

                        if isinstance(eris.ovvv, type(None)):
                            chnk_size = adc.chnk_size
                            if chnk_size > nocc:
                                chnk_size = nocc
                            a = 0
                            for p in range(0,nocc,chnk_size):

                                ka = kconserv[kw, kd, kl]
                                eris_ovvv = dfadc.get_ovvv_df(
                                    adc, eris.Lov[kl,kd], eris.Lvv[kw,ka], p,
                                    chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                                k = eris_ovvv.shape[0]
                                s1 -= 0.5*lib.einsum('lwd,ldwa->a',
                                                     temp_s_a_1[a:a+k],eris_ovvv,optimize=True)
                                s1 -= 0.5*lib.einsum('lwd,ldwa->a',
                                                     temp_t2_r2_2[a:a+k],eris_ovvv,optimize=True)
                                del eris_ovvv

                                eris_ovvv = dfadc.get_ovvv_df(
                                    adc, eris.Lov[kl,ka], eris.Lvv[kw,kd], p,
                                    chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                                s1 += 0.5*lib.einsum('lwd,lawd->a',
                                                     temp_s_a_1[a:a+k],eris_ovvv,optimize=True)
                                s1 += 0.5*lib.einsum('lwd,lawd->a',
                                                     temp_t2_r2_3[a:a+k],eris_ovvv,optimize=True)
                                del eris_ovvv
                                a += k

                        else :
                            ka = kconserv[kw, kd, kl]
                            eris_ovvv = eris.ovvv[:]
                            s1 -= 0.5*lib.einsum('lwd,ldwa->a',temp_s_a_1,
                                                 eris_ovvv[kl,kd,kw],optimize=True)
                            s1 += 0.5*lib.einsum('lwd,lawd->a',temp_s_a_1,
                                                 eris_ovvv[kl,ka,kw],optimize=True)

                            s1 -= 0.5*lib.einsum('lwd,ldwa->a',temp_t2_r2_2,
                                                 eris_ovvv[kl,kd,kw],optimize=True)
                            s1 += 0.5*lib.einsum('lwd,lawd->a',temp_t2_r2_3,
                                                 eris_ovvv[kl,ka,kw],optimize=True)
                            del eris_ovvv

                        del temp_s_a_1
                        del temp_t2_r2_2
                        del temp_t2_r2_3

            for kx in range(nkpts):
                for ky in range(nkpts):
                    ki = kconserv[ky,kshift,kx]
                    for kl in range(nkpts):
                        temp = np.zeros((nocc,nvir,nvir),dtype=eris.oooo.dtype)
                        temp_1_1 = np.zeros((nocc,nvir,nvir),dtype=eris.oooo.dtype)
                        temp_2_1 = np.zeros((nocc,nvir,nvir),dtype=eris.oooo.dtype)
                        if isinstance(eris.ovvv, type(None)):
                            chnk_size = adc.chnk_size
                            if chnk_size > nocc:
                                chnk_size = nocc
                            a = 0
                            for p in range(0,nocc,chnk_size):
                                kd = kconserv[kl, kshift, kx]
                                kb = kconserv[kl,kd,kx]

                                eris_ovvv = dfadc.get_ovvv_df(
                                    adc, eris.Lov[kl,kd], eris.Lvv[kx,kb], p,
                                    chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                                k = eris_ovvv.shape[0]
                                temp_1_1[a:a+k] += lib.einsum('ldxb,b->lxd',
                                                              eris_ovvv,r1,optimize=True)
                                temp_2_1[a:a+k] += lib.einsum('ldxb,b->lxd',
                                                              eris_ovvv,r1,optimize=True)
                                del eris_ovvv

                                eris_ovvv = dfadc.get_ovvv_df(
                                    adc, eris.Lov[kl,kb], eris.Lvv[kx,kd], p,
                                    chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                                temp_1_1[a:a+k] -= lib.einsum('lbxd,b->lxd',
                                                              eris_ovvv,r1,optimize=True)
                                del eris_ovvv

                                kd = kconserv[kl, kshift, ky]
                                kb = kconserv[ky,kd,kl]
                                eris_ovvv = dfadc.get_ovvv_df(
                                    adc, eris.Lov[kl,kb], eris.Lvv[ky,kd], p,
                                    chnk_size).reshape(-1,nvir,nvir,nvir)/nkpts
                                temp[a:a+k] -= lib.einsum('lbyd,b->lyd', eris_ovvv,r1,optimize=True)
                                del eris_ovvv
                                a += k
                        else :
                            kd = kconserv[ki, ky, kl]
                            kb = kconserv[kl,kd,kx]
                            eris_ovvv = eris.ovvv[:]
                            temp_1_1 += lib.einsum('ldxb,b->lxd',
                                                   eris_ovvv[kl,kd,kx],r1,optimize=True)
                            temp_1_1 -= lib.einsum('lbxd,b->lxd',
                                                   eris_ovvv[kl,kb,kx],r1,optimize=True)
                            temp_2_1 += lib.einsum('ldxb,b->lxd',
                                                   eris_ovvv[kl,kd,kx],r1,optimize=True)

                            kd = kconserv[kl, kshift, ky]
                            kb = kconserv[ky,kd,kl]
                            temp -= lib.einsum('lbyd,b->lyd', eris_ovvv[kl,kb,ky],r1,optimize=True)
                            del eris_ovvv

                        kd = kconserv[kl,ky,ki]
                        s2[ki,kx]  += lib.einsum('lxd,lidy->ixy',temp_1_1.conj(),
                                                 t2_1[kl,ki,kd].conj(),optimize=True)
                        s2[ki,kx]  += lib.einsum('lxd,ilyd->ixy',temp_2_1.conj(),
                                                 t2_1[ki,kl,ky].conj(),optimize=True)
                        s2[ki,kx]  -= lib.einsum('lxd,ildy->ixy',temp_2_1.conj(),
                                                 t2_1[ki,kl,kd].conj(),optimize=True)

                        kd = kconserv[kl,kx,ki]
                        s2[ki,kx]  += lib.einsum('lyd,lixd->ixy',temp.conj(),
                                                 t2_1[kl,ki,kx].conj(),optimize=True)
                        del temp
                        del temp_1_1
                        del temp_2_1

        s2 = s2.reshape(-1)
        s = np.hstack((s1,s2))
        del s1
        del s2

        cput0 = log.timer_debug1("completed sigma vector calculation", *cput0)

        return s
    return sigma_


def get_trans_moments(adc,kshift):

    nmo  = adc.nmo
    T = []
    for orb in range(nmo):
        T_a = get_trans_moments_orbital(adc,orb,kshift)
        T.append(T_a)

    T = np.array(T)
    return T


def get_trans_moments_orbital(adc, orb, kshift):

    if adc.method not in ("adc(2)", "adc(2)-x", "adc(3)"):
        raise NotImplementedError(adc.method)

    method = adc.method
    nkpts = adc.nkpts
    nocc = adc.nocc
    nvir = adc.nmo - adc.nocc
    t2_1 = adc.t2[0]

    idn_vir = np.identity(nvir)

    T1 = np.zeros((nvir),dtype=np.complex128)
    T2 = np.zeros((nkpts,nkpts,nocc,nvir,nvir), dtype=np.complex128)

######## ADC(2) 1h part  ############################################

    if orb < nocc:

        if (adc.approx_trans_moments is False or adc.method == "adc(3)"):
            t1_2 = adc.t1[0]
            T1 -= t1_2[kshift][orb,:]

        for kj in range(nkpts):
            for ka in range(nkpts):
                kb = adc.khelper.kconserv[kj, ka, kshift]
                ki = adc.khelper.kconserv[ka, kj, kb]

                t2_1_t= t2_1[ki,kj,ka].transpose(1,0,2,3)
                T2[kj,ka] -= t2_1_t[:,orb,:,:].conj()

    else:

        T1 += idn_vir[(orb-nocc), :]
        for kk in range(nkpts):
            for kc in range(nkpts):
                kl = adc.khelper.kconserv[kc, kk, kshift]
                ka = adc.khelper.kconserv[kc, kl, kk]
                T1 -= 0.25* \
                    lib.einsum('klc,klac->a',t2_1[kk,kl,kshift][:,:,
                               (orb-nocc),:], t2_1[kk,kl,ka].conj(), optimize=True)
                T1 -= 0.25* \
                    lib.einsum('lkc,lkac->a',t2_1[kl,kk,kshift][:,:,
                               (orb-nocc),:], t2_1[kl,kk,ka].conj(), optimize=True)

                T1 -= 0.25* \
                    lib.einsum('klc,klac->a',t2_1[kk,kl,kshift][:,:,
                               (orb-nocc),:], t2_1[kk,kl,ka].conj(), optimize=True)
                T1 += 0.25* \
                    lib.einsum('lkc,klac->a',t2_1[kl,kk,kshift][:,:,
                               (orb-nocc),:], t2_1[kk,kl,ka].conj(), optimize=True)
                T1 += 0.25* \
                    lib.einsum('klc,lkac->a',t2_1[kk,kl,kshift][:,:,
                               (orb-nocc),:], t2_1[kl,kk,ka].conj(), optimize=True)
                T1 -= 0.25* \
                    lib.einsum('lkc,lkac->a',t2_1[kl,kk,kshift][:,:,
                               (orb-nocc),:], t2_1[kl,kk,ka].conj(), optimize=True)

######### ADC(3) 2p-1h  part  ############################################

    if (adc.method == "adc(2)-x" and adc.approx_trans_moments is False) or (adc.method == "adc(3)"):

        t2_2 = adc.t2[1]

        if orb < nocc:

            for kj in range(nkpts):
                for ka in range(nkpts):
                    kb = adc.khelper.kconserv[kj, ka, kshift]
                    ki = adc.khelper.kconserv[ka, kj, kb]

                    t2_2_t = t2_2[ki,kj,ka].conj().transpose(1,0,2,3)

                    T2[kj,ka] -= t2_2_t[:,orb,:,:].conj()


########### ADC(3) 1p part  ############################################

    if(method=='adc(3)'):
        if orb < nocc:
            for kk in range(nkpts):
                for kc in range(nkpts):
                    ka = adc.khelper.kconserv[kk, kc, kshift]
                    T1 += 0.5*lib.einsum('kac,ck->a',
                                         t2_1[kk,kshift,kc][:,orb,:,:], t1_2[kc].T,optimize=True)
                    T1 -= 0.5*lib.einsum('kac,ck->a',
                                         t2_1[kshift,kk,ka][orb,:,:,:], t1_2[kc].T,optimize=True)
                    T1 -= 0.5*lib.einsum('kac,ck->a',
                                         t2_1[kshift,kk,ka][orb,:,:,:], t1_2[kc].T,optimize=True)

        else:
            for kk in range(nkpts):
                for kc in range(nkpts):
                    kl = adc.khelper.kconserv[kk, kc, kshift]
                    ka = adc.khelper.kconserv[kl, kc, kk]

                    T1 -= 0.25* \
                        lib.einsum('klc,klac->a',t2_1[kk,kl,kshift][:,:,
                                   (orb-nocc),:], t2_2[kk,kl,ka].conj(), optimize=True)
                    T1 -= 0.25* \
                        lib.einsum('lkc,lkac->a',t2_1[kl,kk,kshift][:,:,
                                   (orb-nocc),:], t2_2[kl,kk,ka].conj(), optimize=True)
                    T1 -= 0.25* \
                        lib.einsum('klc,klac->a',t2_1[kk,kl,kshift][:,:,
                                   (orb-nocc),:], t2_2[kk,kl,ka].conj(), optimize=True)
                    T1 += 0.25* \
                        lib.einsum('klc,lkac->a',t2_1[kk,kl,kshift][:,:,
                                   (orb-nocc),:], t2_2[kl,kk,ka].conj(), optimize=True)
                    T1 += 0.25* \
                        lib.einsum('lkc,klac->a',t2_1[kl,kk,kshift][:,:,
                                   (orb-nocc),:], t2_2[kk,kl,ka].conj(), optimize=True)
                    T1 -= 0.25* \
                        lib.einsum('lkc,lkac->a',t2_1[kl,kk,kshift][:,:,
                                   (orb-nocc),:], t2_2[kl,kk,ka].conj(), optimize=True)

                    T1 -= 0.25* \
                        lib.einsum('klac,klc->a',t2_1[kk,kl,ka].conj(),
                                   t2_2[kk,kl,kshift][:,:,(orb-nocc),:],optimize=True)
                    T1 -= 0.25* \
                        lib.einsum('lkac,lkc->a',t2_1[kl,kk,ka].conj(),
                                   t2_2[kl,kk,kshift][:,:,(orb-nocc),:],optimize=True)
                    T1 -= 0.25* \
                        lib.einsum('klac,klc->a',t2_1[kk,kl,ka].conj(),
                                   t2_2[kk,kl,kshift][:,:,(orb-nocc),:],optimize=True)
                    T1 += 0.25* \
                        lib.einsum('klac,lkc->a',t2_1[kk,kl,ka].conj(),
                                   t2_2[kl,kk,kshift][:,:,(orb-nocc),:],optimize=True)
                    T1 += 0.25* \
                        lib.einsum('lkac,klc->a',t2_1[kl,kk,ka].conj(),
                                   t2_2[kk,kl,kshift][:,:,(orb-nocc),:],optimize=True)
                    T1 -= 0.25* \
                        lib.einsum('lkac,lkc->a',t2_1[kl,kk,ka].conj(),
                                   t2_2[kl,kk,kshift][:,:,(orb-nocc),:],optimize=True)

        del t2_2
    del t2_1

    for ka in range(nkpts):
        for kb in range(nkpts):
            ki = adc.khelper.kconserv[kb,kshift, ka]
            T2[ki,ka] += T2[ki,ka] - T2[ki,kb].transpose(0,2,1)

    T2 = T2.reshape(-1)
    T = np.hstack((T1,T2))

    return T


def renormalize_eigenvectors(adc, kshift, U, nroots=1):

    nkpts = adc.nkpts
    nocc = adc.t2[0].shape[3]
    nvir = adc.nmo - adc.nocc
    n_singles = nvir

    for I in range(U.shape[1]):
        U1 = U[:n_singles,I]
        U2 = U[n_singles:,I].reshape(nkpts,nkpts,nocc,nvir,nvir)
        UdotU = np.dot(U1.conj().ravel(),U1.ravel())
        for ka in range(nkpts):
            for kb in range(nkpts):
                ki = adc.khelper.kconserv[kb,kshift, ka]
                UdotU +=  2.*np.dot(U2[ki,ka].conj().ravel(), U2[ki,ka].ravel()) - \
                                    np.dot(U2[ki,ka].conj().ravel(),
                                           U2[ki,kb].transpose(0,2,1).ravel())
        U[:,I] /= np.sqrt(UdotU)

    U = U.reshape(-1,nroots)

    return U


def get_properties(adc, kshift, U, nroots=1):

    #Transition moments
    T = adc.get_trans_moments(kshift)

    #Spectroscopic amplitudes
    U = adc.renormalize_eigenvectors(kshift,U,nroots)
    X = np.dot(T, U).reshape(-1, nroots)

    #Spectroscopic factors
    P = 2.0*lib.einsum("pi,pi->i", X.conj(), X)
    P = P.real

    return P,X


class RADCEA(kadc_rhf.RADC):
    '''restricted ADC for EA energies and spectroscopic amplitudes

    Attributes:
        verbose : int
            Print level.  Default value equals to :class:`Mole.verbose`
        max_memory : float or int
            Allowed memory in MB.  Default value equals to :class:`Mole.max_memory`
        incore_complete : bool
            Avoid all I/O. Default is False.
        method : string
            nth-order ADC method. Options are : ADC(2), ADC(2)-X, ADC(3). Default is ADC(2).
        conv_tol : float
            Convergence threshold for Davidson iterations.  Default is 1e-12.
        max_cycle : int
            Number of Davidson iterations.  Default is 50.
        max_space : int
            Space size to hold trial vectors for Davidson iterative diagonalization.  Default is 12.

    Kwargs:
        nroots : int
            Number of roots (eigenvalues) requested. Default value is 1.

            >>> myadc = adc.RADC(mf).run()
            >>> myadcip = adc.RADC(myadc).run()

    Saved results

        e_ea : float or list of floats
            EA energy (eigenvalue). For nroots = 1, it is a single float number. If nroots > 1,
            it is a list of floats for the lowest nroots eigenvalues.
        v_ea : array
            Eigenvectors for each EA transition.
        p_ea : float
            Spectroscopic amplitudes for each EA transition.
    '''

    def __init__(self, adc):
        self.verbose = adc.verbose
        self.stdout = adc.stdout
        self.max_memory = adc.max_memory
        self.max_space = adc.max_space
        self.max_cycle = adc.max_cycle
        self.conv_tol  = adc.conv_tol
        self.tol_residual  = adc.tol_residual
        self.t1 = adc.t1
        self.t2 = adc.t2
        self.method = adc.method
        self.method_type = adc.method_type
        self._scf = adc._scf
        self.compute_properties = adc.compute_properties
        self.approx_trans_moments = adc.approx_trans_moments

        self.kpts = adc._scf.kpts
        self.exxdiv = adc.exxdiv
        self.verbose = adc.verbose
        self.max_memory = adc.max_memory
        self.method = adc.method

        self.khelper = adc.khelper
        self.cell = adc.cell
        self.mo_coeff = adc.mo_coeff
        self.mo_occ = adc.mo_occ
        self.frozen = adc.frozen

        self._nocc = adc._nocc
        self._nmo = adc._nmo
        self._nvir = adc._nvir
        self.nkop_chk = adc.nkop_chk
        self.kop_npick = adc.kop_npick

        self.t2 = adc.t2
        self.e_corr = adc.e_corr
        self.mo_energy = adc.mo_energy
        self.imds = adc.imds
        self.chnk_size = adc.chnk_size

        keys = set(('tol_residual','conv_tol', 'e_corr', 'method', 'mo_coeff', 'mo_energy_b',
                   'max_memory', 't1', 'mo_energy_a', 'max_space', 't2', 'max_cycle'))

        self._keys = set(self.__dict__.keys()).union(keys)

    kernel = kadc_rhf.kernel
    get_imds = get_imds
    get_diag = get_diag
    matvec = matvec
    vector_size = vector_size
    get_trans_moments = get_trans_moments
    renormalize_eigenvectors = renormalize_eigenvectors
    get_properties = get_properties

    def get_init_guess(self, nroots=1, diag=None, ascending=True):
        if diag is None :
            diag = self.get_diag()
        idx = None
        dtype = getattr(diag, 'dtype', np.complex128)
        if ascending:
            idx = np.argsort(diag)
        else:
            idx = np.argsort(diag)[::-1]
        guess = np.zeros((diag.shape[0], nroots), dtype=dtype)
        min_shape = min(diag.shape[0], nroots)
        guess[:min_shape,:min_shape] = np.identity(min_shape)
        g = np.zeros((diag.shape[0], nroots), dtype=dtype)
        g[idx] = guess.copy()
        guess = []
        for p in range(g.shape[1]):
            guess.append(g[:,p])
        return guess

    def gen_matvec(self,kshift,imds=None, eris=None):
        if imds is None:
            imds = self.get_imds(eris)
        diag = self.get_diag(kshift,imds,eris)
        matvec = self.matvec(kshift, imds, eris)
        return matvec, diag


def ea_contract_r_vvvv(adc,r2,vvvv,ka,kb,kc):

    nocc = r2.shape[0]
    nvir = r2.shape[1]
    nkpts = adc.nkpts
    kconserv = adc.khelper.kconserv

    kd = kconserv[ka, kc, kb]
    r2 = np.ascontiguousarray(r2.reshape(nocc,-1))
    r2_vvvv = np.zeros((nvir,nvir,nocc),dtype=r2.dtype)
    chnk_size = adc.chnk_size
    if chnk_size > nvir:
        chnk_size = nvir

    a = 0
    if isinstance(vvvv, np.ndarray):
        vv1 = vvvv[ka,kc]
        vv2 = vvvv[kb,kd]
        for p in range(0,nvir,chnk_size):
            vvvv_p = dfadc.get_vvvv_df(adc, vv1, vv2, p, chnk_size)/nkpts
            k = vvvv_p.shape[0]
            vvvv_p = vvvv_p.reshape(-1,nvir*nvir)
            r2_vvvv[a:a+k] += np.dot(vvvv_p.conj(),r2.T).reshape(-1,nvir,nocc)
            del vvvv_p
            a += k
    else :
        for p in range(0,nvir,chnk_size):
            vvvv_p = vvvv[ka,kb,kc][p:p+chnk_size].reshape(-1,nvir*nvir)
            k = vvvv_p.shape[0]
            r2_vvvv[a:a+k] += np.dot(vvvv_p.conj(),r2.T).reshape(-1,nvir,nocc)
            del vvvv_p
            a += k

    r2_vvvv = np.ascontiguousarray(r2_vvvv.transpose(2,0,1))

    return r2_vvvv



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Relativistic Dirac-Hartree-Fock
"""


import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import _vhf
from pyscf.grad import rhf as rhf_grad


def grad_elec(mf_grad, mo_energy=None, mo_coeff=None, mo_occ=None, atmlst=None):
    mf = mf_grad.base
    mol = mf_grad.mol
    if mo_energy is None: mo_energy = mf.mo_energy
    if mo_occ is None:    mo_occ = mf.mo_occ
    if mo_coeff is None:  mo_coeff = mf.mo_coeff
    log = logger.Logger(mf_grad.stdout, mf_grad.verbose)

    hcore_deriv = mf_grad.hcore_generator(mol)
    s1 = mf_grad.get_ovlp(mol)
    dm0 = mf.make_rdm1(mf.mo_coeff, mf.mo_occ)
    n2c = dm0.shape[0] // 2

    t0 = (logger.process_clock(), logger.perf_counter())
    log.debug('Compute Gradients of NR Hartree-Fock Coulomb repulsion')
    vhf = mf_grad.get_veff(mol, dm0)
    log.timer('gradients of 2e part', *t0)

    dme0 = mf_grad.make_rdm1e(mf.mo_energy, mf.mo_coeff, mf.mo_occ)

    if atmlst is None:
        atmlst = range(mol.natm)
    aoslices = mol.aoslice_2c_by_atom()
    de = numpy.zeros((len(atmlst),3))
    for k, ia in enumerate(atmlst):
        shl0, shl1, p0, p1 = aoslices[ia]
        h1ao = hcore_deriv(ia)
        de[k] += numpy.einsum('xij,ji->x', h1ao, dm0).real
# large components
        de[k] +=(numpy.einsum('xij,ji->x', vhf[:,p0:p1], dm0[:,p0:p1]) +
                 numpy.einsum('xji,ji->x', vhf[:,p0:p1].conj(), dm0[p0:p1])).real
        de[k] -=(numpy.einsum('xij,ji->x', s1[:,p0:p1], dme0[:,p0:p1]) +
                 numpy.einsum('xji,ji->x', s1[:,p0:p1].conj(), dme0[p0:p1])).real
# small components
        p0 += n2c
        p1 += n2c
        de[k] +=(numpy.einsum('xij,ji->x', vhf[:,p0:p1], dm0[:,p0:p1]) +
                 numpy.einsum('xji,ji->x', vhf[:,p0:p1].conj(), dm0[p0:p1])).real
        de[k] -=(numpy.einsum('xij,ji->x', s1[:,p0:p1], dme0[:,p0:p1]) +
                 numpy.einsum('xji,ji->x', s1[:,p0:p1].conj(), dme0[p0:p1])).real
    log.debug('gradients of electronic part')
    log.debug(str(de))
    return de

grad_nuc = rhf_grad.grad_nuc

def get_hcore(mol):
    n2c = mol.nao_2c()
    n4c = n2c * 2
    c = lib.param.LIGHT_SPEED

    t  = mol.intor('int1e_ipkin_spinor', comp=3)
    vn = mol.intor('int1e_ipnuc_spinor', comp=3)
    wn = mol.intor('int1e_ipspnucsp_spinor', comp=3)
    h1e = numpy.zeros((3,n4c,n4c), numpy.complex128)
    h1e[:,:n2c,:n2c] = vn
    h1e[:,n2c:,:n2c] = t
    h1e[:,:n2c,n2c:] = t
    h1e[:,n2c:,n2c:] = wn * (.25/c**2) - t
    return -h1e

def get_ovlp(mol):
    n2c = mol.nao_2c()
    n4c = n2c * 2
    c = lib.param.LIGHT_SPEED

    s  = mol.intor('int1e_ipovlp_spinor', comp=3)
    t  = mol.intor('int1e_ipkin_spinor', comp=3)
    s1e = numpy.zeros((3,n4c,n4c), numpy.complex128)
    s1e[:,:n2c,:n2c] = s
    s1e[:,n2c:,n2c:] = t * (.5/c**2)
    return -s1e

make_rdm1e = rhf_grad.make_rdm1e

def get_coulomb_hf(mol, dm, level='SSSS'):
    '''Dirac-Hartree-Fock Coulomb repulsion'''
    if level.upper() == 'LLLL':
        logger.info(mol, 'Compute Gradients: (LL|LL)')
        vj, vk = _call_vhf1_llll(mol, dm)
#L2SL the response of the large and small components on the large component density
#LS2L the response of the large component on the L+S density
#NOSS just exclude SSSS
#TODO    elif level.upper() == 'LS2L':
#TODO        logger.info(mol, 'Compute Gradients: (LL|LL) + (SS|dLL)')
#TODO        vj, vk = scf.hf.get_vj_vk(pycint.rkb_vhf_coul_grad_ls2l_o1, mol, dm)
#TODO    elif level.upper() == 'L2SL':
#TODO        logger.info(mol, 'Compute Gradients: (LL|LL) + (dSS|LL)')
#TODO        vj, vk = scf.hf.get_vj_vk(pycint.rkb_vhf_coul_grad_l2sl_o1, mol, dm)
#TODO    elif level.upper() == 'NOSS':
#TODO        logger.info(mol, 'Compute Gradients: (LL|LL) + (dSS|LL) + (SS|dLL)')
#TODO        vj, vk = scf.hf.get_vj_vk(pycint.rkb_vhf_coul_grad_xss_o1, mol, dm)
    else:
        logger.info(mol, 'Compute Gradients: (LL|LL) + (SS|LL) + (SS|SS)')
        vj, vk = _call_vhf1(mol, dm)
    return -(vj - vk)
get_veff = get_coulomb_hf


class GradientsMixin(rhf_grad.GradientsMixin):
    '''
    Basic nuclear gradient functions for 4C relativistic methods
    '''
    def get_hcore(self, mol=None):
        if mol is None: mol = self.mol
        return get_hcore(mol)

    def hcore_generator(self, mol):
        aoslices = mol.aoslice_2c_by_atom()
        h1 = self.get_hcore(mol)
        n2c = mol.nao_2c()
        n4c = n2c * 2
        c = lib.param.LIGHT_SPEED
        def hcore_deriv(atm_id):
            shl0, shl1, p0, p1 = aoslices[atm_id]
            with mol.with_rinv_at_nucleus(atm_id):
                z = -mol.atom_charge(atm_id)
                vn = z * mol.intor('int1e_iprinv_spinor', comp=3)
                wn = z * mol.intor('int1e_ipsprinvsp_spinor', comp=3)

            v = numpy.zeros((3,n4c,n4c), numpy.complex128)
            v[:,:n2c,:n2c] = vn
            v[:,n2c:,n2c:] = wn * (.25/c**2)
            v[:,p0:p1]         += h1[:,p0:p1]
            v[:,n2c+p0:n2c+p1] += h1[:,n2c+p0:n2c+p1]
            return v + v.conj().transpose(0,2,1)
        return hcore_deriv

    def get_ovlp(self, mol=None):
        if mol is None: mol = self.mol
        return get_ovlp(mol)


class Gradients(GradientsMixin):
    '''Unrestricted Dirac-Hartree-Fock gradients'''
    def __init__(self, scf_method):
        GradientsMixin.__init__(self, scf_method)
        if scf_method.with_ssss:
            self.level = 'SSSS'
        else:
            #self.level = 'NOSS'
            #self.level = 'LLLL'
            raise NotImplementedError
        self._keys = self._keys.union(['level'])

    def get_veff(self, mol, dm):
        return get_coulomb_hf(mol, dm, level=self.level)

    def make_rdm1e(self, mo_energy=None, mo_coeff=None, mo_occ=None):
        if mo_energy is None: mo_energy = self.base.mo_energy
        if mo_coeff is None: mo_coeff = self.base.mo_coeff
        if mo_occ is None: mo_occ = self.base.mo_occ
        return make_rdm1e(mo_energy, mo_coeff, mo_occ)

    grad_elec = grad_elec

    def extra_force(self, atom_id, envs):
        '''Hook for extra contributions in analytical gradients.

        Contributions like the response of auxiliary basis in density fitting
        method, the grid response in DFT numerical integration can be put in
        this function.
        '''
        return 0

    def kernel(self, mo_energy=None, mo_coeff=None, mo_occ=None, atmlst=None):
        cput0 = (logger.process_clock(), logger.perf_counter())
        if mo_energy is None: mo_energy = self.base.mo_energy
        if mo_coeff is None: mo_coeff = self.base.mo_coeff
        if mo_occ is None: mo_occ = self.base.mo_occ
        if atmlst is None:
            atmlst = self.atmlst
        else:
            self.atmlst = atmlst

        if self.verbose >= logger.WARN:
            self.check_sanity()
        if self.verbose >= logger.INFO:
            self.dump_flags()

        de = self.grad_elec(mo_energy, mo_coeff, mo_occ, atmlst)
        self.de = de + self.grad_nuc(atmlst=atmlst)
        if self.mol.symmetry:
            self.de = self.symmetrize(self.de, atmlst)
        logger.timer(self, 'SCF gradients', *cput0)
        self._finalize()
        return self.de

    as_scanner = rhf_grad.as_scanner

Grad = Gradients

from pyscf import scf
scf.dhf.UHF.Gradients = lib.class_as_method(Gradients)


def _call_vhf1_llll(mol, dm):
    n2c = dm.shape[0] // 2
    dmll = dm[:n2c,:n2c].copy()
    vj = numpy.zeros((3,n2c*2,n2c*2), dtype=numpy.complex128)
    vk = numpy.zeros((3,n2c*2,n2c*2), dtype=numpy.complex128)
    vj[:,:n2c,:n2c], vk[:,:n2c,:n2c] = \
            _vhf.rdirect_mapdm('int2e_ip1_spinor', 's2kl',
                               ('lk->s1ij', 'jk->s1il'), dmll, 3,
                               mol._atm, mol._bas, mol._env)
    return vj, vk

def _call_vhf1(mol, dm):
    c1 = .5 / lib.param.LIGHT_SPEED
    n2c = dm.shape[0] // 2
    dmll = dm[:n2c,:n2c].copy()
    dmls = dm[:n2c,n2c:].copy()
    dmsl = dm[n2c:,:n2c].copy()
    dmss = dm[n2c:,n2c:].copy()
    vj = numpy.zeros((3,n2c*2,n2c*2), dtype=numpy.complex128)
    vk = numpy.zeros((3,n2c*2,n2c*2), dtype=numpy.complex128)
    vj[:,:n2c,:n2c], vk[:,:n2c,:n2c] = \
            _vhf.rdirect_mapdm('int2e_ip1_spinor', 's2kl',
                               ('lk->s1ij', 'jk->s1il'), dmll, 3,
                               mol._atm, mol._bas, mol._env)
    vj[:,n2c:,n2c:], vk[:,n2c:,n2c:] = \
            _vhf.rdirect_mapdm('int2e_ipspsp1spsp2_spinor', 's2kl',
                               ('lk->s1ij', 'jk->s1il'), dmss, 3,
                               mol._atm, mol._bas, mol._env) * c1**4
    vx = _vhf.rdirect_bindm('int2e_ipspsp1_spinor', 's2kl',
                            ('lk->s1ij', 'jk->s1il'), (dmll, dmsl), 3,
                            mol._atm, mol._bas, mol._env) * c1**2
    vj[:,n2c:,n2c:] += vx[0]
    vk[:,n2c:,:n2c] += vx[1]
    vx = _vhf.rdirect_bindm('int2e_ip1spsp2_spinor', 's2kl',
                            ('lk->s1ij', 'jk->s1il'), (dmss, dmls), 3,
                            mol._atm, mol._bas, mol._env) * c1**2
    vj[:,:n2c,:n2c] += vx[0]
    vk[:,:n2c,n2c:] += vx[1]
    return vj, vk


if __name__ == "__main__":
    from pyscf import gto
    from pyscf import scf
    from pyscf import lib

    with lib.light_speed(30):
        h2o = gto.Mole()
        h2o.verbose = 0
        h2o.output = None#"out_h2o"
        h2o.atom = [
            ["O" , (0. , 0.     , 0.)],
            [1   , (0. , -0.757 , 0.587)],
            [1   , (0. , 0.757  , 0.587)] ]
        h2o.basis = {"H": '6-31g',
                     "O": '6-31g',}
        h2o.build()
        method = scf.dhf.UHF(h2o).run()
        g = method.Gradients().kernel()
        print(g)

        ms = method.as_scanner()
        h2o.set_geom_([["O" , (0. , 0.     ,-0.001)],
                       [1   , (0. , -0.757 , 0.587)],
                       [1   , (0. , 0.757  , 0.587)]], unit='Ang')
        e1 = ms(h2o)
        h2o.set_geom_([["O" , (0. , 0.     , 0.001)],
                       [1   , (0. , -0.757 , 0.587)],
                       [1   , (0. , 0.757  , 0.587)]], unit='Ang')
        e2 = ms(h2o)
        print(g[0,2], (e2-e1)/0.002*lib.param.BOHR)



#!/usr/bin/env python
#
# This code was copied from the data generation program of Tencent Alchemy
# project (https://github.com/tencent-alchemy).

#
# Copyright 2019 Tencent America LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Thermochemistry analysis.

Ref:
    psi4/psi4/driver/qcdb/vib.py
    http://gaussian.com/vib/
'''

from functools import reduce
import numpy
from pyscf import lib
from pyscf.data import nist

LINDEP_THRESHOLD = 1e-7


def harmonic_analysis(mol, hess, exclude_trans=True, exclude_rot=True,
                      imaginary_freq=True, mass=None):
    '''Each column is one mode

    imaginary_freq (boolean): save imaginary_freq as complex number (if True)
    or negative real number (if False)
    '''
    if mass is None:
        mass = mol.atom_mass_list(isotope_avg=True)
    results = {}
    atom_coords = mol.atom_coords()
    mass_center = numpy.einsum('z,zx->x', mass, atom_coords) / mass.sum()
    atom_coords = atom_coords - mass_center
    natm = atom_coords.shape[0]

    mass_hess = numpy.einsum('pqxy,p,q->pqxy', hess, mass**-.5, mass**-.5)
    h = mass_hess.transpose(0,2,1,3).reshape(natm*3,natm*3)

    TR = _get_TR(mass, atom_coords)
    TRspace = []
    if exclude_trans:
        TRspace.append(TR[:3])

    if exclude_rot:
        rot_const = rotation_const(mass, atom_coords)
        rotor_type = _get_rotor_type(rot_const)
        if rotor_type == 'ATOM':
            pass
        elif rotor_type == 'LINEAR':  # linear molecule
            TRspace.append(TR[3:5])
        else:
            TRspace.append(TR[3:])

    if TRspace:
        TRspace = numpy.vstack(TRspace)
        q, r = numpy.linalg.qr(TRspace.T)
        P = numpy.eye(natm * 3) - q.dot(q.T)
        w, v = numpy.linalg.eigh(P)
        bvec = v[:,w > LINDEP_THRESHOLD]
        h = reduce(numpy.dot, (bvec.T, h, bvec))
        force_const_au, mode = numpy.linalg.eigh(h)
        mode = bvec.dot(mode)
    else:
        force_const_au, mode = numpy.linalg.eigh(h)

    freq_au = numpy.lib.scimath.sqrt(force_const_au)
    results['freq_error'] = numpy.count_nonzero(freq_au.imag > 0)
    if not imaginary_freq and numpy.iscomplexobj(freq_au):
        # save imaginary frequency as negative frequency
        freq_au = freq_au.real - abs(freq_au.imag)

    results['freq_au'] = freq_au
    au2hz = (nist.HARTREE2J / (nist.ATOMIC_MASS * nist.BOHR_SI**2))**.5 / (2 * numpy.pi)
    results['freq_wavenumber'] = freq_au * au2hz / nist.LIGHT_SPEED_SI * 1e-2

    norm_mode = numpy.einsum('z,zri->izr', mass**-.5, mode.reshape(natm,3,-1))
    results['norm_mode'] = norm_mode
    reduced_mass = 1./numpy.einsum('izr,izr->i', norm_mode, norm_mode)
    results['reduced_mass'] = reduced_mass

    # https://en.wikipedia.org/wiki/Vibrational_temperature
    results['vib_temperature'] = freq_au * au2hz * nist.PLANCK / nist.BOLTZMANN

    # force constants
    dyne = 1e-2 * nist.HARTREE2J / nist.BOHR_SI**2
    results['force_const_au'] = force_const_au
    results['force_const_dyne'] = reduced_mass * force_const_au * dyne  #cm^-1/a0^2

    #TODO: IR intensity
    return results

def rotation_const(mass, atom_coords, unit='GHz'):
    '''Rotational constants to characterize rotational spectra

    Kwargs:
        unit (string) : One of GHz, wavenumber
    '''
    mass_center = numpy.einsum('z,zr->r', mass, atom_coords) / mass.sum()
    r = atom_coords - mass_center
    im = numpy.einsum('z,zr,zs->rs', mass, r, r)
    im = numpy.eye(3) * im.trace() - im
    e = numpy.sort(numpy.linalg.eigvalsh(im))

    unit_im = nist.ATOMIC_MASS * (nist.BOHR_SI)**2
    unit_hz = nist.HBAR / (4 * numpy.pi * unit_im)
    with numpy.errstate(divide='ignore'):
        if unit.lower() == 'ghz':
            e = unit_hz / e * 1e-9
        elif unit.lower() == 'wavenumber':
            e = unit_hz / e / nist.LIGHT_SPEED_SI * 1e-2
        else:
            raise RuntimeError('Unsupported unit ' + unit)
    return e


def thermo(model, freq, temperature=298.15, pressure=101325):
    mol = model.mol
    atom_coords = mol.atom_coords()
    mass = mol.atom_mass_list(isotope_avg=True)
    mass_center = numpy.einsum('z,zx->x', mass, atom_coords) / mass.sum()
    atom_coords = atom_coords - mass_center

    kB = nist.BOLTZMANN
    h = nist.PLANCK
    # c = nist.LIGHT_SPEED_SI
    # beta = 1. / (kB * temperature)
    R_Eh = kB*nist.AVOGADRO / (nist.HARTREE2J * nist.AVOGADRO)

    results = {}
    results['temperature'] = (temperature, 'K')
    results['pressure'] = (pressure, 'Pa')

    E0 = model.e_tot
    results['E0'] = (E0, 'Eh')

    # Electronic part
    results['S_elec' ] = (R_Eh * numpy.log(mol.multiplicity), 'Eh/K')
    results['Cv_elec'] = results['Cp_elec'] = (0, 'Eh/K')
    results['E_elec' ] = results['H_elec' ] = (E0, 'Eh')

    # Translational part. See also https://cccbdb.nist.gov/thermo.asp for the
    # partition function q_trans
    mass_tot = mass.sum() * nist.ATOMIC_MASS
    q_trans = ((2.0 * numpy.pi * mass_tot * kB * temperature / h**2)**1.5
               * kB * temperature / pressure)
    results['S_trans' ] = (R_Eh * (2.5 + numpy.log(q_trans)), 'Eh/K')
    results['Cv_trans'] = (1.5 * R_Eh, 'Eh/K')
    results['Cp_trans'] = (2.5 * R_Eh, 'Eh/K')
    results['E_trans' ] = (1.5 * R_Eh * temperature, 'Eh')
    results['H_trans' ] = (2.5 * R_Eh * temperature, 'Eh')

    # Rotational part
    rot_const = rotation_const(mass, atom_coords, 'GHz')
    results['rot_const'] = (rot_const, 'GHz')
    rotor_type = _get_rotor_type(rot_const)

    sym_number = rotational_symmetry_number(mol)
    results['sym_number'] = (sym_number, '')

    # partition function q_rot (https://cccbdb.nist.gov/thermo.asp)
    if rotor_type == 'ATOM':
        results['S_rot' ] = (0, 'Eh/K')
        results['Cv_rot'] = results['Cp_rot'] = (0, 'Eh/K')
        results['E_rot' ] = results['H_rot' ] = (0, 'Eh')
    elif rotor_type == 'LINEAR':
        B = rot_const[1] * 1e9
        q_rot = kB * temperature / (sym_number * h * B)
        results['S_rot' ] = (R_Eh * (1 + numpy.log(q_rot)), 'Eh/K')
        results['Cv_rot'] = results['Cp_rot'] = (R_Eh, 'Eh/K')
        results['E_rot' ] = results['H_rot' ] = (R_Eh * temperature, 'Eh')
    else:
        ABC = rot_const * 1e9
        q_rot = ((kB*temperature/h)**1.5 * numpy.pi**.5
                 / (sym_number * numpy.prod(ABC)**.5))
        results['S_rot' ] = (R_Eh * (1.5 + numpy.log(q_rot)), 'Eh/K')
        results['Cv_rot'] = results['Cp_rot'] = (1.5 * R_Eh, 'Eh/K')
        results['E_rot' ] = results['H_rot' ] = (1.5 * R_Eh * temperature, 'Eh')

    # Vibrational part.
    au2hz = (nist.HARTREE2J / (nist.ATOMIC_MASS * nist.BOHR_SI**2))**.5 / (2 * numpy.pi)
    idx = freq.real > 0
    vib_temperature = freq.real[idx] * au2hz * h / kB
    # reduced_temperature
    rt = vib_temperature / max(1e-14, temperature)
    e = numpy.exp(-rt)

    ZPE = R_Eh * .5 * vib_temperature.sum()
    results['ZPE'] = (ZPE, 'Eh')

    results['S_vib' ] = (R_Eh * (rt*e/(1-e) - numpy.log(1-e)).sum(), 'Eh/K')
    results['Cv_vib'] = results['Cp_vib'] = (R_Eh * (e * rt**2/(1-e)**2).sum(), 'Eh/K')
    results['E_vib' ] = results['H_vib' ] = \
            (ZPE + R_Eh * temperature * (rt * e / (1-e)).sum(), 'Eh')

    results['G_elec' ] = (results['H_elec' ][0] - temperature * results['S_elec' ][0], 'Eh')
    results['G_trans'] = (results['H_trans'][0] - temperature * results['S_trans'][0], 'Eh')
    results['G_rot'  ] = (results['H_rot'  ][0] - temperature * results['S_rot'  ][0], 'Eh')
    results['G_vib'  ] = (results['H_vib'  ][0] - temperature * results['S_vib'  ][0], 'Eh')

    def _sum(f):
        keys = ('elec', 'trans', 'rot', 'vib')
        return sum(results.get(f+'_'+key, (0,))[0] for key in keys)
    results['S_tot' ] = (_sum('S' ), 'Eh/K')
    results['Cv_tot'] = (_sum('Cv'), 'Eh/K')
    results['Cp_tot'] = (_sum('Cp'), 'Eh/K')
    results['E_0K' ]  = (E0 + ZPE, 'Eh')
    results['E_tot' ] = (_sum('E'), 'Eh')
    results['H_tot' ] = (_sum('H'), 'Eh')
    results['G_tot' ] = (_sum('G'), 'Eh')

    return results


def _get_TR(mass, coords):
    '''Translational mode and rotational mode'''
    mass_center = numpy.einsum('z,zx->x', mass, coords) / mass.sum()
    coords = coords - mass_center
    massp = mass ** .5

    # translational mode
    Tx = numpy.einsum('m,x->mx', massp, [1, 0, 0])
    Ty = numpy.einsum('m,x->mx', massp, [0, 1, 0])
    Tz = numpy.einsum('m,x->mx', massp, [0, 0, 1])

    im = numpy.einsum('m,mx,my->xy', mass, coords, coords)
    im = numpy.eye(3) * im.trace() - im
    w, paxes = numpy.linalg.eigh(im)
    # make the z-axis be the rotation vector with the smallest moment of inertia
    w = w[::-1]
    paxes = paxes[:,::-1]
    ex, ey, ez = paxes.T

    # rotational mode
    coords_in_rot_frame = coords.dot(paxes)
    cx, cy, cz = coords_in_rot_frame.T
    Rx = massp[:,None] * (cy[:,None] * ez - cz[:,None] * ey)
    Ry = massp[:,None] * (cz[:,None] * ex - cx[:,None] * ez)
    Rz = massp[:,None] * (cx[:,None] * ey - cy[:,None] * ex)

    return (Tx.ravel(), Ty.ravel(), Tz.ravel(),
            Rx.ravel(), Ry.ravel(), Rz.ravel())


def _get_rotor_type(rot_const):
    if numpy.all(rot_const > 1e8):
        rotor_type = 'ATOM'
    elif rot_const[0] > 1e8 and (rot_const[1] - rot_const[2] < 1e-3):
        rotor_type = 'LINEAR'
    else:
        rotor_type = 'REGULAR'
    return rotor_type


def rotational_symmetry_number(mol):
    '''Number of unique orientations of the rigid molecule that only
    interchange identical atoms.

    Source http://cccbdb.nist.gov/thermo.asp (search "symmetry number")
    '''
    from pyscf import symm
    group = symm.detect_symm(mol._atom)[0]

    if group in ['SO3', 'C1', 'Ci', 'Cs', 'Coov']:
        sigma = 1
    elif group == 'Dooh':
        sigma = 2
    elif group in ['T', 'Td']:
        sigma = 12
    elif group == 'Oh':
        sigma = 24
    elif group == 'Ih':
        sigma = 60
    elif group[0] == 'C': # 'Cn', 'Cnv', 'Cnh'
        sigma = int(''.join([x for x in group if x.isdigit()]))
    elif group[0] == 'D': # 'Dn', 'Dnd', 'Dnh'
        sigma = 2 * int(''.join([x for x in group if x.isdigit()]))
    elif group[0] == 'S': # 'Sn'
        sigma = int(''.join([x for x in group if x.isdigit()])) / 2
    else:
        raise RuntimeError("symmetry group: " + group)
    return sigma



def dump_thermo(mol, results):
    dump = mol.stdout.write
    dump('temperature %.4f [%s]\n' % results['temperature'])
    dump('pressure %.2f [%s]\n' % results['pressure'])
    dump('Rotational constants [%s] %.5f %.5f %.5f\n'
         % ((results['rot_const'][1],) + tuple(results['rot_const'][0])))
    dump('Symmetry number %d\n' % results['sym_number'][0])
    dump('Zero-point energy (ZPE) %.5f [Eh]   %.3f [J/mol]\n'
         % (results['ZPE'][0], results['ZPE'][0] * nist.HARTREE2J * nist.AVOGADRO))

    keys = ('tot', 'elec', 'trans', 'rot', 'vib')
    dump('                    %s\n' % ' '.join('%10s'%x for x in keys))
    def convert(f, keys, unit):
        if 'Eh' in unit:
            conv = nist.HARTREE2J * nist.AVOGADRO
        else:
            conv = 1
        return ' '.join('%10.3f'%(results.get(f+'_'+key, (0,))[0]*conv) for key in keys)
    def write(title, f):
        tot, unit = results[f+'_tot']
        msg = convert(f, keys, unit)
        unit = unit.replace('Eh', 'J/mol')
        s = '%s [%s]' % (title, unit)
        dump('%-20s %s\n' % (s, msg))
    write('Entropy', 'S')
    write('Cv', 'Cv')
    write('Cp', 'Cp')

    dump('%-28s               %s\n'
         % ('Internal energy [J/mol]', convert('E', keys[2:], 'Eh')))
    dump('%-22s %.5f  %.5f\n'
         % ('Internal energy [Eh]', results['E_tot'][0], results['E0'][0]))
    dump('%-28s               %s\n'
         % ('Enthalpy [J/mol]', convert('H', keys[2:], 'Eh')))
    dump('%-22s %.5f\n'
         % ('Enthalpy [Eh]', results['H_tot'][0]))
    dump('%-28s               %s\n'
         % ('Gibbs free energy [J/mol]', convert('G', keys[2:], 'Eh')))
    dump('%-22s %.5f\n'
         % ('Gibbs free energy [Eh]', results['G_tot'][0]))

def dump_normal_mode(mol, results):
    dump = mol.stdout.write
    freq_wn = results['freq_wavenumber']
    idx = freq_wn.real > 0
    freq_wn = freq_wn.real[idx]
    nfreq = freq_wn.size

    r_mass = results['reduced_mass'].real[idx]
    force = results['force_const_dyne'].real[idx]
    vib_t = results['vib_temperature'].real[idx]
    mode = results['norm_mode'].real[idx]
    symbols = [mol.atom_symbol(i) for i in range(mol.natm)]

    def inline(q, col0, col1):
        return ''.join('%20.4f' % q[i] for i in range(col0, col1))
    def mode_inline(row, col0, col1):
        return '  '.join('%6.2f%6.2f%6.2f' % (mode[i,row,0], mode[i,row,1], mode[i,row,2])
                         for i in range(col0, col1))

    for col0, col1 in lib.prange(0, nfreq, 3):
        dump('Mode              %s\n' % ''.join('%20d'%i for i in range(col0,col1)))
        dump('Irrep\n')
        dump('Freq [cm^-1]          %s\n' % inline(freq_wn, col0, col1))
        dump('Reduced mass [au]     %s\n' % inline(r_mass, col0, col1))
        dump('Force const [Dyne/A]  %s\n' % inline(force, col0, col1))
        dump('Char temp [K]         %s\n' % inline(vib_t, col0, col1))
        #dump('IR\n')
        #dump('Raman\n')
        dump('Normal mode            %s\n' % ('       x     y     z'*(col1-col0)))
        for j, at in enumerate(symbols):
            dump('    %4d%4s               %s\n' % (j, at, mode_inline(j, col0, col1)))


if __name__ == '__main__':
    from pyscf import gto
    from pyscf import hessian
    mol = gto.M(atom='O 0 0 0; H 0 .757 .587; H 0 -.757 .587')

    mass = mol.atom_mass_list(isotope_avg=True)
    r = mol.atom_coords() - numpy.random.random((1,3))
    print(rotation_const(mass, r, 'GHz'))
    print(rotation_const(mass[1:], r[1:], 'GHz'))
    print(rotation_const(mass[2:], r[2:], 'GHz'))

    mf = mol.apply('HF').run()
    hess = hessian.RHF(mf).kernel()
    results = harmonic_analysis(mol, hess)
    dump_normal_mode(mol, results)

    results = thermo(mf, results['freq_au'], 298.15, 101325)
    dump_thermo(mol, results)



# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyscf.pbc.cc import ccsd
from pyscf.pbc import scf

def KRCCSD(mf, frozen=None, mo_coeff=None, mo_occ=None):
    from pyscf.pbc.mpicc import kccsd_rhf
    mf = scf.addons.convert_to_rhf(mf)
    return kccsd_rhf.RCCSD(mf, frozen)

KCCSD = KRCCSD
